#include <Wire.h>
#include "LiquidCrystal_I2C.h"
#include "DinoController.h"
#include "WeedController.h"
#include "CactiController.h"
#include "Controller.h"

// RetroLCD.com - 2020
// https://forum.arduino.cc/index.php?topic=117045.0
// Set the LCD address to 0x27 for a 16 chars and 2 line display
LiquidCrystal_I2C lcd(0x27, 16, 2);

char Screen[32];
int screen_x = 16;
int screen_y = 2;
int start_timer;
int end_timer;
float frame_time;
float total_time;
int player_x;
int max_x = 500;
int points = 0;

float weed_x;
float weed_y;

DinoController dino;
WeedController weed;
CactiController cacti;

float cacti_x[10];
int cacti_sprite[10];
int cacti_count = 10;

void setup() {
  Serial.begin(9600);

  randomSeed(analogRead(0));

  // initialize the LCD
  lcd.begin();
  lcd.clear();
  lcd.backlight();
  lcd.setCursor(0, 0);

  dino.Init();
  weed.Init();
  Controller::Init();
  player_x = 0;

  lcd.createChar(6, cacti.frame1);
  lcd.createChar(7, cacti.frame2);

  total_time = 0;
  weed_x = random(200, 1000);
  int j;
  for (j = 0; j < cacti_count; j++) {
    cacti_x[j] = random(17, max_x);
    cacti_sprite[j] = random(6, 8);
  }
}

void ClearScreen() {
  for (int j = 0; j < 32; j++) {
    Screen[j] = ' ';
  }
}

void PlotCHAR(int x, int y, char c) {
  if (x + y * screen_x >= 32) {
    return;
  }
  if (x < 0) {
    return;
  }
  if (y < 0) {
    return;
  }
  if (x > 15) {
    return;
  }
  if (y > 1) {
    return;
  }
  Screen[x + y * screen_x] = c;
}

void RenderScreen() {
  int x, y;
  char line[17];
  byte point_char[16];

  String point_disp = "Pts:" + String(points);
  point_disp.getBytes(point_char, 16);

  for (x = 0; x < 16; x++) {
    PlotCHAR(x + 6, 0, (char)point_char[x]);
  }

  lcd.setCursor(0, 0);

  for (y = 0; y < screen_y; y++) {
    for (x = 0; x < screen_x; x++) {
      line[x] = Screen[x + y * screen_x];
    }
    line[16] = 0;
    lcd.setCursor(0, y);
    lcd.print(line);
  }
}

void HandleInput()
{
  if (dino.dino_state != DINO_STATE_JUMP
      && dino.dino_state != DINO_STATE_CROUCH) {
    if (Controller::IsPressedAgain(CONTROLLER_BUTTON_A)) {
      dino.Jump();
      Controller::MarkUnreleased(CONTROLLER_BUTTON_A);
    } else {
      if (Controller::IsPressedAgain(CONTROLLER_BUTTON_B)) {
        dino.Crouch();
        Controller::MarkUnreleased(CONTROLLER_BUTTON_B);
      } else {
        if (Controller::IsPressedAgain(CONTROLLER_BUTTON_LEFT)) {
          if (player_x > 0) {
            player_x--;
          }
          Controller::MarkUnreleased(CONTROLLER_BUTTON_LEFT);
        } else {
          if (Controller::IsPressedAgain(CONTROLLER_BUTTON_RIGHT)) {
            if (player_x < 4) {
              player_x++;
            }
            Controller::MarkUnreleased(CONTROLLER_BUTTON_RIGHT);
          }
        }
      }
    }
  }
}

void HandleSpriteChange()
{
  if (weed.weed_sprite_change) {
    lcd.createChar(5, weed.frame);
    weed.weed_sprite_change = false;
  }

  if (dino.dino_sprite_change) {
    // NOTE: do not use createChar(0, ...), it confuses the Ardino
    // http://forum.arduino.cc/index.php?topic=74666.0
    lcd.createChar(1, dino.dinoTL);
    lcd.createChar(2, dino.dinoTR);
    lcd.createChar(3, dino.dinoBL);
    lcd.createChar(4, dino.dinoBR);
    dino.dino_sprite_change = false;
  }
}

void HandleSpriteUpdates()
{
  weed_x -= frame_time * 4.0;
  if (weed_x < 0) {
    points += 10;
    weed_x = random(200, 1000);
  }
  weed_y = (int)(sin(total_time * 6.0 / 3.14) + 1);

  int j;
  for (j = 0; j < cacti_count; j++) {
    cacti_x[j] -= frame_time * 4.0;
    if (cacti_x[j] < 0) {
      points++;
      cacti_x[j] = random(17, max_x);
      cacti_sprite[j] = random(6, 8);
    }
  }
}

void HandleCollision()
{
  switch (dino.dino_state) {
    case DINO_STATE_JUMP:
      if (weed_y == 0)
      {
        if (weed_x >= player_x && weed_x <= player_x + 2) {
          weed_x = random(200, 1000);
          points -= 5;
          dino.Hit();
        }
      }
      break;
    case DINO_STATE_CROUCH:
      if (weed_y == 1)
      {
        if (weed_x >= player_x && weed_x <= player_x + 2) {
          weed_x = random(200, 1000);
          points -= 5;
          dino.Hit();
        }
      }
      break;
    default:
      if (weed_x >= player_x && weed_x <= player_x + 1) {
        weed_x = random(200, 1000);
        points -= 5;
        dino.Hit();
      }
      break;
  }

  if (dino.dino_state != DINO_STATE_JUMP) {
    int j;
    for (j = 0; j < cacti_count; j++) {
      if (player_x == (int)cacti_x[j]) {
        cacti_x[j] = random(17, max_x);
        cacti_sprite[j] = random(6, 8);
        points--;
        dino.Hit();
        break;
      }

    }
  }
  if (points < 0) {
    points = 0;
  }
}

void loop() {
  start_timer = millis();

  Controller::ReadButtons();

  HandleInput();
  HandleSpriteChange();
  HandleSpriteUpdates();
  HandleCollision();


  ClearScreen();
  PlotCHAR((int) weed_x, (int) weed_y, 5);
  switch (dino.dino_state) {
    case DINO_STATE_JUMP:
      PlotCHAR(player_x + 0, 0, 1);
      PlotCHAR(player_x + 1, 0, 2);
      PlotCHAR(player_x + 2, 0, 3);
      break;
    case DINO_STATE_CROUCH:
      PlotCHAR(player_x + 0, 1, 1);
      PlotCHAR(player_x + 1, 1, 2);
      PlotCHAR(player_x + 2, 1, 3);
      break;
    default:
      PlotCHAR(player_x + 0, 0, 1);
      PlotCHAR(player_x + 1, 0, 2);
      PlotCHAR(player_x + 0, 1, 3);
      PlotCHAR(player_x + 1, 1, 4);
      break;
  }

  int j;
  for (j = 0; j < cacti_count; j++) {
    PlotCHAR((int)cacti_x[j], 1, cacti_sprite[j]);
  }

  RenderScreen();

  end_timer = millis();
  frame_time = (float) (end_timer - start_timer) / 1000.0;
  dino.Update(frame_time);
  weed.Update(frame_time);

  total_time += frame_time;
  total_time = ((total_time / 60.0) - (int) (total_time / 60.0)) * 60.0;

}
